/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

// SPDX-FileCopyrightText: 2022 - 2025 Kohei Yoshida
//
// SPDX-License-Identifier: MIT

#pragma once

#ifndef MDDS_MTV_USE_STANDARD_ELEMENT_BLOCKS
#define MDDS_MTV_USE_STANDARD_ELEMENT_BLOCKS 1
#endif

/* vim:set shiftwidth=4 softtabstop=4 expandtab: */
